@@ -21,17 +21,16 @@ import random
 import sys
 from argparse import ArgumentParser, Namespace
 from dataclasses import dataclass
-from logging import Logger
 from operator import itemgetter
 from typing import Any, Dict, List, Optional
 
 import gym
+import orjson
 from gym.wrappers import TimeLimit
 
 from gym_saturation.envs import SaturationEnv
 from gym_saturation.envs.saturation_env import STATE_DIFF_UPDATED
 from gym_saturation.logic_ops.utils import clause_length
-from gym_saturation.parsing.json_grammar import dict_to_clause
 
 
 class BaseAgent:
@@ -93,14 +92,18 @@ class SizeAgent(BaseAgent):
         :param info: an info dict (parm of environment response)
         :returns:
         """
+        parsed_state_diff = tuple(
+            (index, orjson.loads(clause))
+            for index, clause in info[STATE_DIFF_UPDATED].items()
+        )
         self._state.update(
             {
-                index: clause_length(dict_to_clause(clause))
-                for index, clause in info[STATE_DIFF_UPDATED].items()
+                index: clause_length(clause)
+                for index, clause in parsed_state_diff
                 if not clause["processed"]
             }
         )
-        for index, clause in info[STATE_DIFF_UPDATED].items():
+        for index, clause in parsed_state_diff:
             if clause["processed"]:
                 self._state.pop(index)
 
@@ -133,7 +136,9 @@ class AgeAgent(BaseAgent):
         return min(
             [
                 i
-                for i, clause in enumerate(observation["real_obs"])
+                for i, clause in enumerate(
+                    map(orjson.loads, observation["real_obs"])
+                )
                 if not clause["processed"]
             ]
         )
@@ -186,15 +191,15 @@ class RandomAgent(BaseAgent):
         return random.choice(
             [
                 i
-                for i, clause in enumerate(observation["real_obs"])
+                for i, clause in enumerate(
+                    map(orjson.loads, observation["real_obs"])
+                )
                 if not clause["processed"]
             ]
         )
 
 
-def episode(
-    env: SaturationEnv, agent: BaseAgent, logger: Optional[Logger] = None
-) -> Transition:
+def episode(env: SaturationEnv, agent: BaseAgent) -> Transition:
     """
     tries to solve the problem and logs the clauses
 
@@ -234,8 +239,6 @@ def episode(
     :param env: a `gym_saturation` environment
     :param agent: an initialized agent. Must have `get_action` method
     :param problem_filename: the name of a problem file
-    :param logger: where to log the episode memory;
-        if ``None`` then nothing is logged
     :returns: the last transition
     """
     env_state, reward, done = env.reset(), 0.0, False
@@ -254,8 +257,6 @@ def episode(
             done,
             info,
         )
-        if logger is not None:
-            logger.info(transition)
         env_state = observation
     return transition
 
