@@ -15,11 +15,13 @@
 Saturation Environment
 =======================
 """
+import dataclasses
 import os
 import random
-from typing import Dict, List, Tuple
+from typing import Dict, List, Set, Tuple
 
 import numpy as np
+import orjson
 from gym import Env, spaces
 
 from gym_saturation.clause_space import ClauseSpace
@@ -31,12 +33,10 @@ from gym_saturation.logic_ops.reflexivity_resolution import (
 )
 from gym_saturation.logic_ops.resolution import all_possible_resolvents
 from gym_saturation.logic_ops.utils import (
-    clause_in_a_list,
     is_tautology,
     reduce_to_proof,
     reindex_variables,
 )
-from gym_saturation.parsing.json_grammar import clause_to_dict
 from gym_saturation.parsing.tptp_parser import TPTPParser
 
 STATE_DIFF_UPDATED = "state_diff_updated"
@@ -84,8 +84,8 @@ class SaturationEnv(Env):
     ``ansi`` mode returns a JSON representation of the state
     it should be more easily parsable than TPTP, although less human-friendly
 
-    >>> env.render("ansi")[:73]
-    "[{'class': 'Clause', 'literals': [{'class': 'Literal', 'negated': False, "
+    >>> env.render("ansi") # doctest: +ELLIPSIS
+    '...{"literals":[{"negated":false,"atom":{"name":"this_is_a_test_case","arguments":[{"name":...
 
     other modes are not implemented yet
 
@@ -100,7 +100,8 @@ class SaturationEnv(Env):
 
     ``info`` dict contains the state diff, for example
 
-    >>> info["state_diff_updated"][0]["processed"]
+    >>> import json
+    >>> json.loads(info["state_diff_updated"][0])["processed"]
     True
 
     repeating actions is not allowed
@@ -134,7 +135,8 @@ class SaturationEnv(Env):
     ):
         super().__init__()
         self.problem_list = problem_list
-        self._state: List[Clause] = []
+        self._state: Tuple[Clause, ...] = ()
+        self._state_set: Set[Tuple[bytes, ...]] = set()
         self.action_space = spaces.Discrete(max_clauses)
         self.observation_space = spaces.Dict(
             {
@@ -143,41 +145,59 @@ class SaturationEnv(Env):
             }
         )
 
-    def _init_clauses(self):
+    def _init_clauses(self) -> Tuple[Clause, ...]:
         problem = random.choice(self.problem_list)
         tptp_folder = os.path.join(os.path.dirname(problem), "..", "..")
         with open(problem, "r", encoding="utf-8") as problem_file:
             problem_text = problem_file.read()
-        clauses = TPTPParser().parse(problem_text, tptp_folder)
-        for clause in clauses:
-            clause.birth_step = 0
-            clause.inference_parents = []
-            clause.inference_rule = None
-            clause.processed = False
-        return clauses
+        parsed_clauses = TPTPParser().parse(problem_text, tptp_folder)
+        return tuple(
+            dataclasses.replace(
+                clause,
+                birth_step=0,
+                inference_parents=(),
+                inference_rule=None,
+                processed=False,
+            )
+            for clause in parsed_clauses
+        )
 
     def reset(self) -> dict:
         self._state = reindex_variables(self._init_clauses(), "X")
+        self._state_set = set(
+            map(
+                lambda clause: tuple(
+                    sorted(map(orjson.dumps, clause.literals))
+                ),
+                self._state,
+            )
+        )
         return self.state
 
-    def _add_to_state(self, new_clauses: List[Clause]) -> None:
+    def _add_to_state(self, new_clauses: Tuple[Clause, ...]) -> None:
         birth_step = 1 + max(
             [getattr(clause, "birth_step", 0) for clause in self._state]
         )
         for clause in new_clauses:
             if not is_tautology(clause):
-                if not clause_in_a_list(clause, self._state):
-                    clause.birth_step = birth_step
-                    clause.processed = False
-                    self._state.append(clause)
+                sorted_literals = tuple(
+                    sorted(map(orjson.dumps, clause.literals))
+                )
+                if sorted_literals not in self._state_set:
+                    self._state = self._state + (
+                        dataclasses.replace(
+                            clause, birth_step=birth_step, processed=False
+                        ),
+                    )
+                    self._state_set.add(sorted_literals)
 
-    def _do_deductions(self, action: int) -> Dict[int, Clause]:
+    def _do_deductions(self, action: int) -> Dict[int, bytes]:
         state_len_before = len(self._state)
         given_clause = self._state[action]
         if not given_clause.processed:
-            unprocessed_clauses = [
+            unprocessed_clauses = tuple(
                 clause for clause in self._state if clause.processed
-            ]
+            )
             self._add_to_state(
                 all_possible_resolvents(
                     unprocessed_clauses,
@@ -200,27 +220,35 @@ class SaturationEnv(Env):
                     given_clause,
                 )
             )
-        self._state[action].processed = True
+        self._state = (
+            self._state[:action]
+            + (dataclasses.replace(self._state[action], processed=True),)
+            + self._state[action + 1 :]
+        )
         return dict(
             [
-                (i + state_len_before, clause_to_dict(clause))
+                (i + state_len_before, orjson.dumps(clause))
                 for i, clause in enumerate(self._state[state_len_before:])
             ]
-            + [(action, clause_to_dict(self._state[action]))]
+            + [(action, orjson.dumps(self._state[action]))]
         )
 
     def step(self, action: int) -> Tuple[dict, float, bool, dict]:
         if self._state[action].processed:
             raise ValueError(f"action {action} is not valid")
-        if self._state[action].literals == []:
-            self._state[action].processed = True
+        if self._state[action].literals == ():
+            self._state = (
+                self._state[:action]
+                + (dataclasses.replace(self._state[action], processed=True),)
+                + self._state[action + 1 :]
+            )
             return (
                 self.state,
                 1.0,
                 True,
                 {
                     STATE_DIFF_UPDATED: {
-                        action: clause_to_dict(self._state[action])
+                        action: orjson.dumps(self._state[action])
                     }
                 },
             )
@@ -258,7 +286,7 @@ class SaturationEnv(Env):
                 "consider increasing `max_clauses` parameter of `__init__`"
             )
         return {
-            "real_obs": clause_to_dict(self._state),
+            "real_obs": [orjson.dumps(clause) for clause in self._state],
             "action_mask": (
                 np.array(
                     [
